import java.io.*;
import java.util.*;

public class Maze {

    private int SearchDelay;
    
    public class Room implements BinaryHeap.HeapElement{
        private int xpos, ypos;
        private VisitedState state;
        private RoomList neighbors;
	private Room prevOnPath;
	private int priority;
	private int aStarHeapPos;
	private int distFromStart;



        Room(int x, int y){
            xpos = x;
            ypos = y;
            neighbors = new RoomList();
	    reInit();
	} 
        
	public void reInit()
	{
            state = VisitedState.UNSEEN;
	    aStarHeapPos = -1;
	    prevOnPath = null;
	    priority = Integer.MAX_VALUE;
	}
		
	public Room getPrevOnPath()	         { return prevOnPath; }
	public void setPrevOnPath(Room room) { prevOnPath = room; }

	public int  getDistFromStart()	         { return distFromStart; }
	public void setDistFromStart(int value)  { distFromStart = value; }

	public int  getHeapPos()      { return aStarHeapPos; }
	public void setHeapPos(int x) {	aStarHeapPos = x; }

	public int  getPriority()             { return priority; }
	

	public boolean decreasePriority(int pri) 
	{ 
	    boolean decreasedPriority = priority > pri;
	    priority = (decreasedPriority)?pri:priority;
	    return decreasedPriority;
	}


	public int  comparePriority (BinaryHeap.HeapElement x)  { return priority - x.getPriority(); }
		
	public int getX() { return xpos; }
	public int getY() { return ypos; }

	public VisitedState getState() { return state; }
	
	public VisitedState getIDSState( boolean polarity )
	{
		if( (!polarity && aStarHeapPos == 0) || (polarity && aStarHeapPos == 1) )
			return state;
		return VisitedState.UNSEEN;
	}
        void setIDSState(VisitedState s, boolean polarity){
            setState(s);
	    if(polarity) aStarHeapPos = 1;
	    else aStarHeapPos = 0;
        }
	
        RoomList getNeighbors() { return neighbors; }

        boolean Adjacent(Room otherRoom){
            for(RoomNode rn = neighbors.first; rn != null; rn = rn.next)
                if (rn.data == otherRoom) return true;
            return false;
        }

        void setState(VisitedState s){
	    int dummy;
            state = s;
            canvas.repaint();
	    try { Thread.sleep( SearchDelay ); } catch( Exception ex ) {}
	    	    dummy = 0;
	    	    while (dummy < 100) {
			dummy = dummy + 1;
	    }
	}
        
//////////////////////////////////////////////////////////
//////////////
        public boolean equals( Room x )
        {
          if(   x.getX() == this.xpos 
             && x.getY() == this.ypos )
            return true;
          return false;
        }
//////////////
//////////////////////////////////////////////////////////


    }

    public class RoomNode {
        private Room data;
        private RoomNode next;
    
        RoomNode(Room r){
            data = r;
            next = null;
        }
    
	RoomNode(Room r, RoomNode n) {
		data = r;
		next = n;
	}

        Room     getRoom() { return data; }
	RoomNode getNext() { return next; }
    }

    public class RoomList {
        private RoomNode first;

        RoomList(){
            first = null;
        }

        void Insert(Room r){
            first = new RoomNode(r,first);
        }
        RoomNode getFirst(){ return first; }
    }

    // Beginning of class Maze proper
    private Room start, end;
    private int width, height, cycles;
    private Room [][] rooms;
    private MazeCanvas canvas;

///////////////////////////////////////////////////////////
//////////////
    public Room getStart() { return start; }
    public Room getEnd()   { return end; }
    
    public boolean isEnd( Room x ) { return end.equals( x ); }
    
    public int getHeight() { return height; }
    public int getWidth()  { return width; }
    public Room[][] getRooms() { return rooms; }

    

    public void setCanvas ( MazeCanvas canvas ){
        this.canvas = canvas;   
    }
    
    public void reInitRooms()
    {
      for (int i=1; i<=width; i++)
          for (int j=1; j<=height; j++)
              rooms[i][j].reInit();
    } 


//////////////
//////////////////////////////////////////////////////////

    Room getRoomAt(int x, int y){ return rooms[x][y]; }


///////////////////////////////////////////////////////////
//////////////
    Maze( int width, int height, int cycles, int delay )
    {

	//System.out.println("starting Maze"); 

	this.SearchDelay = delay;
      // WARNING: Do not create a maze of size 1,1
      this.width = width;
      this.height = height;
      int maxCycles = width*height - width - height + 1;
      cycles = (cycles>maxCycles)?maxCycles:cycles;
      this.cycles = cycles;  
      
      canvas = null;
      
      int row, col;
      // Number of adjacencies in maze
      int numAdjs = (2*(width*height)-width-height);
      // Adjacency lists
      int [] adjList1 = new int[ numAdjs ];
      int [] adjList2 = new int[ numAdjs ];

      // Populate adjacency list (adjacent room numbers)
      int l = 0;
      for( int k = 0; k<(height*(width-1)); k+=(width-1) )
      {
        for( int w = 0; w<(width-1); w++ )
        {
          adjList1[w+k] = w+k+l;
          adjList2[w+k] = w+k+l+1;
        }
        l++;
      }
      l = 0;
      for( int k = (height*(width-1)); k<numAdjs ; k+=width )
      {
        for( int h = 0; h<width; h++ )
        {
          adjList1[h+k] = h+l;
          adjList2[h+k] = h+width+l;
        }
        l+=width;
      }

      // Populate 2-d array with room objects
      rooms = new Room[width+1][height+1];
      for (int i=1; i<=width; i++)
          for (int j=1; j<=height; j++)
              rooms[i][j] = new Room(i,j);

      // Generate union find object
      UnionFind uf = new UnionFind( width*height );

      // Create and seed random number generator with current time
      Random number = new Random();

      int rm1, rm2;
      int root1, root2;
      int randNum;
      int row1, col1;
      int row2, col2;
      int numCycles = cycles;
      int a=0;
      // While there are unconnected rooms
      for(a = 0; !uf.isOneSet();)
      {
        // Choose a random adjacency
        randNum = number.nextInt( numAdjs-a );
        rm1 = adjList1[ randNum ];
        rm2 = adjList2[ randNum ];
        // Check union/find if the rooms are connected
        root1 = uf.find( rm1 );
        root2 = uf.find( rm2 );
        // If not connected already
        if( root1 != root2 )
        {
          // Remove it from the list by swapping in the last element
          adjList1[randNum] = adjList1[numAdjs-a-1];
          adjList2[randNum] = adjList2[numAdjs-a-1];
          // Union the nodes in the union/find using room numbers
          uf.union( root1, root2 );
          // Translate connected room numbers into x,y coordinates
          row1 = (rm1/width)+1;
          col1 = (rm1%width)+1;
          row2 = (rm2/width)+1;
          col2 = (rm2%width)+1;
          // To the room adj 2-d array, make them neighbors (2-way)
          rooms[col1][row1].getNeighbors().Insert( rooms[col2][row2] );
          rooms[col2][row2].getNeighbors().Insert( rooms[col1][row1] );
 	  a++;
        }
      }
      // Make the cycles
      for(int i = 0, j = cycles; j>0 && numAdjs-a-i>0;i++,j--)
      {
        // Choose a random adjacency
        randNum = number.nextInt( numAdjs-a-i );
        rm1 = adjList1[ randNum ];
        rm2 = adjList2[ randNum ];
        // Remove it from the list by swapping in the last element
        adjList1[randNum] = adjList1[numAdjs-a-i-1];
        adjList2[randNum] = adjList2[numAdjs-a-i-1];
        // Translate connected room numbers into x,y coordinates
        row1 = (rm1/width)+1;
        col1 = (rm1%width)+1;
        row2 = (rm2/width)+1;
        col2 = (rm2%width)+1;
        // To the room adj 2-d array, make them neighbors (2-way)
        rooms[col1][row1].getNeighbors().Insert( rooms[col2][row2] );
        rooms[col2][row2].getNeighbors().Insert( rooms[col1][row1] );
      }
      // Choose a random start and end
      if( width > 2 && height > 2 )
      {
        row1 = number.nextInt( height );
        col1 = number.nextInt( width );
        row2 = number.nextInt( height );
        col2 = number.nextInt( width );
        // Random number generter includes 0
        // Correct to 1
        // Check start coordinates
        if( 0 == row1 )
          row1 = 1;
        if( 0 == col1 )
          col1 = 1;
        // Check end coordinates
        if( 0 == row2 )
          row2 = 1;
        if( 0 == col2 )
          col2 = 1;
        // Random number geneterator might choose start == end, correct
        while( row1 == row2 && col1 == col2 )
        {
          // Get new coordinates
          row2 = number.nextInt( height );
          col2 = number.nextInt( width );
          // Correct if needed
          if( 0 == row2 )
            row2 = 1;
          if( 0 == col2 )
            col2 = 1;
        }
        start = rooms[col1][row1];
        end = rooms[col2][row2];
      }
      else
      {
        // Choose start to be corners b/c the above iteration with random
        // number generator might take a while to find unique start and end locations
        start = rooms[1][1];
        end = rooms[width][height]; 
      }

      // Print to file
      /*
      try
      {
        File f = new File( "mazegen.txt" ); 
        FileOutputStream fos = new FileOutputStream( f );
        PrintStream ps = new PrintStream( fos );
        ps.println( width + " " + height );
        printMaze( fos );
      }
      catch( Exception e )
      {
        System.out.println( "Unable to print generated maze to file." );
      }
      */
      //      System.out.println("ending Maze"); 
    }
//////////////
///////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
//////////////
    public void printMaze( OutputStream os )
    {
      PrintStream out = new PrintStream( os );
      printMaze( out );
    }
//////////////
///////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
//////////////
    private void printMaze(PrintStream out){

	// first, the top wall
	for (int i=1; i<=width; i++)
        {
	    out.print('+');
            out.print('-');
        }
	out.print('+');
	out.print('\n');

	for (int j=1; j<=height; j++) { 
	// line w/walls, rooms, and doors
	    out.print('|');
	    for (int i=1; i<=width; i++) {
		if (start == rooms[i][j])
		    out.print('*');
		else if (end == rooms[i][j])
		    out.print('X');
		else {
		    VisitedState v = rooms[i][j].getState();
		    if (v == VisitedState.UNSEEN)
			out.print(' ');
		    else if (v == VisitedState.VISITED)
			out.print('#');
		    else if (v == VisitedState.ONPATH)
			out.print('.');
		    else if (v == VisitedState.PROCESSING)
			out.print('?');
		}
		if (i < width) {
		    if (rooms[i][j].Adjacent(rooms[i+1][j]))
			out.print(' ');
		    else 
			out.print('|');
		}
	    }
	    out.print('|');
	    out.print('\n');

	    // line w/walls and doors only
	    for (int i=1; i<=width; i++) {
		if (i > 1 && j < height &&
		    rooms[i][j].Adjacent(rooms[i-1][j]) &&
		    rooms[i][j+1].Adjacent(rooms[i-1][j+1]) &&
		    rooms[i-1][j].Adjacent(rooms[i-1][j+1]) &&
		    rooms[i][j+1].Adjacent(rooms[i][j]))
		    out.print(' ');
		else
		    out.print('+');
		if (j < height && rooms[i][j].Adjacent(rooms[i][j+1]))
		    out.print(' ');
		else
		    out.print('-');
	    }
	    out.print('+');
	    out.print('\n');
	}
    }
//////////////
///////////////////////////////////////////////////////////

}
