///////////////////////////////////////////////////////////////////////////////
//
//  Authors:	Jeffrey Hanke (hanke@cs)
//		Jason Dang (jaydang@cs)
//		Chris Baker (crbaker@cs)
//
//  File:	MazeSearch.java
//
//  Abstract:	Implements all search algorithms for MazeRunner project
//
///////////////////////////////////////////////////////////////////////////////

import java.util.*;
import java.io.*;

public class MazeSearch extends Thread
{

  private static final int TIMER_DELAY = 0;
    
  private Maze theMaze;
  private String theSearch;
  private int iPathLength;
  private int iNodesExpanded;
  private PathList thePath;
  private String curSearch;
  private AMazer applet;
  private boolean keepRunning;

  MazeSearch( Maze aMaze, String curSearch, AMazer applet )
  {
    theMaze = aMaze;
    thePath = new PathList();
    theSearch = "No search";
    iPathLength = 0;
    iNodesExpanded = 0;
    this.curSearch = curSearch;
    this.applet = applet;
    keepRunning = true;
  }

  public void stopRunning()
  { 
    keepRunning = false;
  }


  public void printStats()
  {
    System.out.println( theSearch );
    System.out.println( thePath );
    System.out.println( iPathLength );
    System.out.println( iNodesExpanded );
  }

  public void run()
  {
      try { Thread.sleep( TIMER_DELAY ); } catch( Exception ex ) {}
    if( curSearch.equals( "dfs" ) )
      this.depthFirstSearch();
    else if( curSearch.equals( "bfs" ) )
      this.breadthFirstSearch();
    else if( curSearch.equals( "ids" ) )
      this.iterativeDepthFirstSearch();
    else if( curSearch.equals( "best" ) )
      this.bestFirstSearch();
    else if( curSearch.equals( "astar" ) )
      this.aStarSearch();
    if( keepRunning ){
      applet.printText( theSearch + "\n"
                        + thePath + "\n"
                        + iPathLength + "\n"
                        + iNodesExpanded );
      applet.printStats(iPathLength, iNodesExpanded);
    }
  }

  //
  //  Method:	iterativeDepthFirstSearch()
  //
  //  Abstract:	Performs an iterative depth first search on the given maze	
  //
  public void iterativeDepthFirstSearch()
  {
    Queue q = new Queue(); // I use addFirst so the Queue will actually be a stack.

    Maze.RoomNode rn;
    Maze.Room CurRoom = null, AdjRoom = null;
    boolean polarity = false; // The polarity of the IDS iteration
    // Statistical bookeeping
    theSearch = "Iterative Depth-First Search";
    iPathLength = -1; 
    // Main iteration while the queue is not empty
    // and we haven't found the end
    for(int i=0; CurRoom == null || !theMaze.isEnd(CurRoom);i++, q.makeEmpty(), polarity = !polarity)
    {
	    for(CurRoom = theMaze.getStart(), CurRoom.setDistFromStart(0);CurRoom != null && !theMaze.isEnd(CurRoom); CurRoom = (Maze.Room) q.dequeue() )
	    {
                if( !keepRunning )
                  return;
	      	if(CurRoom.getDistFromStart()+1>i) continue;
	        // Iterate over all neighbors
	        for( rn = (CurRoom.getNeighbors()).getFirst();
				 rn!=null; 
				 rn = rn.getNext() )
	        {
	           AdjRoom = rn.getRoom();
	           if( VisitedState.UNSEEN == AdjRoom.getIDSState(polarity) )
	           {
	              AdjRoom.setIDSState( VisitedState.PROCESSING, polarity );
	      	      AdjRoom.setDistFromStart(CurRoom.getDistFromStart()+1);
	              q.addFirst( AdjRoom );
		      AdjRoom.setPrevOnPath(CurRoom);
	           }
	           else if( (VisitedState.VISITED == AdjRoom.getIDSState(polarity) || VisitedState.PROCESSING == AdjRoom.getIDSState(polarity))
		   	    && (CurRoom.getDistFromStart()+1)<AdjRoom.getDistFromStart() )
	           {
	              AdjRoom.setIDSState( VisitedState.PROCESSING, polarity );
	              q.addFirst( AdjRoom );
	      	      AdjRoom.setDistFromStart(CurRoom.getDistFromStart()+1);
		      AdjRoom.setPrevOnPath(CurRoom);
	           }
                   try { Thread.sleep( TIMER_DELAY ); } catch( Exception e ) {}
	        }
	        // Set the current room to visited
	        CurRoom.setIDSState( VisitedState.VISITED, polarity );
	        // Statistical bookeeping
	        iNodesExpanded++;
	    }

    }
	for(;CurRoom != null; CurRoom = CurRoom.getPrevOnPath())
	{
		thePath.addFirst( CurRoom.getX(), CurRoom.getY() );
		iPathLength++;
		CurRoom.setState( VisitedState.ONPATH );
	}
  }
  
  //
  //  Method:	depthFirstSearch()
  //
  //  Abstract:	Performs a depth first search on the given maze	
  //
  public void depthFirstSearch()
  {
    Queue q = new Queue(); // I use addFirst so the Queue will actually be a stack.

    Maze.RoomNode rn;
    Maze.Room CurRoom, AdjRoom;

    // Statistical bookeeping
    theSearch = "Depth-First Search";
    iPathLength = -1; 
    // Main iteration while the queue is not empty
    // and we haven't found the end
    for(CurRoom = theMaze.getStart(),CurRoom.setDistFromStart(0);!theMaze.isEnd(CurRoom) ; CurRoom = (Maze.Room) q.dequeue() )
    {
        if( !keepRunning )
          return;
        // Iterate over all neighbors
        for( rn = (CurRoom.getNeighbors()).getFirst();
			 rn!=null; 
			 rn = rn.getNext() )
        {
           AdjRoom = rn.getRoom();
           if( VisitedState.UNSEEN == AdjRoom.getState() )
           {
              AdjRoom.setState( VisitedState.PROCESSING );
	      AdjRoom.setDistFromStart(CurRoom.getDistFromStart()+1);
              q.addFirst( AdjRoom );
	      AdjRoom.setPrevOnPath(CurRoom);
           }
           try { Thread.sleep( TIMER_DELAY ); } catch( Exception e ) {}

        }
        // Set the current room to visited
        CurRoom.setState( VisitedState.VISITED );
        // Statistical bookeeping
        iNodesExpanded++;
    }
	
	for(;CurRoom != null; CurRoom = CurRoom.getPrevOnPath())
	{
		thePath.addFirst( CurRoom.getX(), CurRoom.getY() );
		iPathLength++;
		CurRoom.setState( VisitedState.ONPATH );
	}
  }
  
  //
  //  Method:	breadthFirstSearch()
  //
  //  Abstract:	Performs a breadth first search on the given maze	
  //
  public void breadthFirstSearch()
  {
    Queue q = new Queue();

    Maze.RoomNode rn;
    Maze.Room CurRoom, AdjRoom;

    // Statistical bookeeping
    theSearch = "Breadth-First Search";
    iPathLength = -1; 
    // Main iteration while the queue is not empty
    // and we haven't found the end
    for(CurRoom = theMaze.getStart();!theMaze.isEnd(CurRoom) ; CurRoom = (Maze.Room) q.dequeue() )
    {
        if( !keepRunning )
          return;
        // Iterate over all neighbors
        for( rn = (CurRoom.getNeighbors()).getFirst();
			 rn!=null; 
			 rn = rn.getNext() )
        {
           AdjRoom = rn.getRoom();
           if( VisitedState.UNSEEN == AdjRoom.getState() )
           {
              AdjRoom.setState( VisitedState.PROCESSING );
	      AdjRoom.setDistFromStart(CurRoom.getDistFromStart()+1);
              q.enqueue( AdjRoom );
	 	AdjRoom.setPrevOnPath(CurRoom);
           }
           try { Thread.sleep( TIMER_DELAY ); } catch( Exception e ) {}
        }
        // Set the current room to visited
        CurRoom.setState( VisitedState.VISITED );
        // Statistical bookeeping
        iNodesExpanded++;
    }
	
	for(;CurRoom != null; CurRoom = CurRoom.getPrevOnPath())
	{
		thePath.addFirst( CurRoom.getX(), CurRoom.getY() );
		iPathLength++;
		CurRoom.setState( VisitedState.ONPATH );
	}
  }
  
  //
  //  Method:	bestFirstSearch()
  //
  //  Abstract:	Performs a best-first search on the given maze	
  //
  public void bestFirstSearch()
  {
    BinaryHeap q = new BinaryHeap();

    Maze.RoomNode rn;
    Maze.Room CurRoom, AdjRoom;
    int endX = theMaze.getEnd().getX(), endY = theMaze.getEnd().getY();
    
    // Statistical bookeeping
    theSearch = "Best-First Search";
    iPathLength = -1; 
    // Main iteration while the queue is not empty
    // and we haven't found the end
    for(CurRoom = theMaze.getStart(),CurRoom.setDistFromStart(0);!theMaze.isEnd(CurRoom) ; CurRoom = (Maze.Room) q.deleteMin() )
    {
        if( !keepRunning )
          return;
        // Iterate over all neighbors
        for( rn = (CurRoom.getNeighbors()).getFirst();
			 rn!=null; 
			 rn = rn.getNext() )
        {
           AdjRoom = rn.getRoom();
           if( VisitedState.UNSEEN == AdjRoom.getState() )
           {
              AdjRoom.setState( VisitedState.PROCESSING );
	      AdjRoom.setDistFromStart(CurRoom.getDistFromStart()+1);
              q.decreaseKeyInsert( AdjRoom, Math.abs(AdjRoom.getX()-endX) + Math.abs(AdjRoom.getY()-endY) );
	 	AdjRoom.setPrevOnPath(CurRoom);
           }
           try { Thread.sleep( TIMER_DELAY ); }catch( Exception e ) {} 
        }
        // Set the current room to visited
        CurRoom.setState( VisitedState.VISITED );
        // Statistical bookeeping
        iNodesExpanded++;
    }
	
	for(;CurRoom != null; CurRoom = CurRoom.getPrevOnPath())
	{
		thePath.addFirst( CurRoom.getX(), CurRoom.getY() );
		iPathLength++;
		CurRoom.setState( VisitedState.ONPATH );
	}
    }
  
  //
  //  Method:	aStarSearch()
  //
  //  Abstract:	Performs an A * search on the given maze	
  //



/////////////////////////////////////////////////

  public void aStarSearch()
  {
    BinaryHeap q = new BinaryHeap();

    Maze.RoomNode rn;
    Maze.Room CurRoom, AdjRoom;
    int endX = theMaze.getEnd().getX(), endY = theMaze.getEnd().getY();
    
    // Statistical bookeeping
    theSearch = "A * Search";
    iPathLength = -1; 
    // Main iteration while the queue is not empty
    // and we haven't found the end
    for(CurRoom = theMaze.getStart(),CurRoom.setDistFromStart(0);!theMaze.isEnd(CurRoom) ; CurRoom = (Maze.Room) q.deleteMin() )
    {
        if( !keepRunning )
          return;
        // Iterate over all neighbors
        for( rn = (CurRoom.getNeighbors()).getFirst();
			 rn!=null; 
			 rn = rn.getNext() )
        {
           AdjRoom = rn.getRoom();
           if( VisitedState.VISITED != AdjRoom.getState() )
           {
              AdjRoom.setState( VisitedState.PROCESSING );
	      if(AdjRoom.decreasePriority((CurRoom.getDistFromStart() + 1 + Math.abs(AdjRoom.getX()-endX) + Math.abs(AdjRoom.getY()-endY))*theMaze.getHeight()*theMaze.getWidth()
	      	                     +AdjRoom.getX()*theMaze.getHeight()+AdjRoom.getY()))
	      {
	          AdjRoom.setDistFromStart(CurRoom.getDistFromStart()+1);
                  q.decreaseKeyInsert( AdjRoom, AdjRoom.getPriority());
	          AdjRoom.setPrevOnPath(CurRoom);
 	      }
              try { Thread.sleep( TIMER_DELAY ); }catch( Exception e ) {}
           }
        }
        // Set the current room to visited
        CurRoom.setState( VisitedState.VISITED );
        // Statistical bookeeping
        iNodesExpanded++;
    }
	
	for(;CurRoom != null; CurRoom = CurRoom.getPrevOnPath())
	{
		thePath.addFirst( CurRoom.getX(), CurRoom.getY() );
		iPathLength++;
		CurRoom.setState( VisitedState.ONPATH );
	}
  }
}
